"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.onPostBuild = exports.createResolvers = exports.createPages = exports.createSchemaCustomization = exports.sourceNodes = exports.pluginOptionsSchema = void 0;
const fs_1 = __importDefault(require("fs"));
const gatsby_graphql_source_toolkit_1 = require("gatsby-graphql-source-toolkit");
const create_pages_1 = require("./helpers/create-pages");
const create_query_executor_1 = require("./helpers/create-query-executor");
const create_sourcing_config_1 = require("./helpers/create-sourcing-config");
const create_translation_query_field_1 = require("./helpers/create-translation-query-field");
const drupal_feeds_1 = require("./helpers/drupal-feeds");
const fetch_node_changes_1 = require("./helpers/fetch-node-changes");
const utils_1 = require("./utils");
const pluginOptionsSchema = ({ Joi, }) => Joi.object({
    drupal_url: Joi.string().uri({ allowRelative: false }).required(),
    graphql_path: Joi.string().uri({ relativeOnly: true }).required(),
    drupal_external_url: Joi.string().uri({ allowRelative: false }),
    auth_user: Joi.string().optional(),
    auth_pass: Joi.string().optional(),
    auth_key: Joi.string().optional(),
    query_concurrency: Joi.number().optional().min(1),
    paginator_page_size: Joi.number().optional().min(2),
});
exports.pluginOptionsSchema = pluginOptionsSchema;
const getForwardedHeaders = (url) => ({
    'X-Forwarded-Proto': url.protocol,
    'X-Forwarded-Host': url.hostname,
    'X-Forwarded-Port': url.port,
});
const sourceNodes = (gatsbyApi, options) => __awaiter(void 0, void 0, void 0, function* () {
    var _a, _b;
    if (!(0, utils_1.validOptions)(options)) {
        return;
    }
    const executor = (0, create_query_executor_1.createQueryExecutor)(Object.assign(Object.assign({}, options), { headers: options.drupal_external_url
            ? getForwardedHeaders(new URL(options.drupal_external_url))
            : undefined }));
    // TODO: Accept configuration and fragment overrides from plugin settings.
    const config = yield (0, create_sourcing_config_1.createSourcingConfig)(gatsbyApi, executor, options);
    const context = (0, gatsby_graphql_source_toolkit_1.createSourcingContext)(config);
    // Source only what was changed. If there is something in cache.
    const lastBuildId = (yield gatsbyApi.cache.get(`LAST_BUILD_ID`)) || -1;
    let currentBuildId = ((_a = gatsbyApi.webhookBody) === null || _a === void 0 ? void 0 : _a.buildId) || -1;
    // If the webhook did not contain a build, we attempt to fetch the
    // latest one from drupal.
    if (currentBuildId === -1) {
        const info = yield executor({
            operationName: 'LatestBuildId',
            variables: {},
            query: `
      query LatestBuildId {
        drupalBuildId
      }
    `,
        });
        currentBuildId = ((_b = info === null || info === void 0 ? void 0 : info.data) === null || _b === void 0 ? void 0 : _b.drupalBuildId) || -1;
    }
    // If the current build is lower than the last one, the CMS has been reset and we
    // need to re-fetch everything. If the two are equal, this is a manual request, in which
    // case we also re-fetch all data. If we don't have a last build id, we can't trust any
    // data that is stored in Gatsby and we have to re-fetch everything anyway.
    if (
    // Current build id is lower than the last one -> out of sync with CMS, re-fetch everything
    currentBuildId < lastBuildId ||
        // No information about a current build in the CMS -> re-fetch everything
        currentBuildId === -1 ||
        // No information about the latest build in Gatsby -> re-fetch everything
        lastBuildId === -1) {
        gatsbyApi.reporter.info(`ℹ️ clearing all nodes.`);
        const feeds = yield (0, drupal_feeds_1.drupalFeeds)(executor);
        for (const feed of feeds) {
            const nodes = gatsbyApi.getNodesByType(`Drupal${feed.typeName}`);
            const events = nodes.map((node) => ({
                remoteTypeName: feed.typeName,
                eventName: 'DELETE',
                remoteId: { id: node.id },
            }));
            (0, gatsby_graphql_source_toolkit_1.deleteNodes)(context, events);
        }
        // If we don't have a last build or the CMS has not information about the
        // latest build, there is no way to detect changes. We have to run a full
        // rebuild.
        gatsbyApi.reporter.info(`ℹ️ sourceNodes will fetch all nodes.`);
        yield (0, gatsby_graphql_source_toolkit_1.sourceAllNodes)(config);
    }
    else {
        gatsbyApi.reporter.info(`Fetching changes between builds ${lastBuildId} and ${currentBuildId}.`);
        const nodeEvents = lastBuildId && currentBuildId
            ? yield (0, fetch_node_changes_1.fetchNodeChanges)(lastBuildId, currentBuildId, gatsbyApi.reporter, context, executor)
            : [];
        yield (0, gatsby_graphql_source_toolkit_1.sourceNodeChanges)(config, { nodeEvents });
    }
    gatsbyApi.reporter.info(`sourced data for build ${currentBuildId}`);
    yield gatsbyApi.cache.set(`LAST_BUILD_ID_TMP`, currentBuildId);
});
exports.sourceNodes = sourceNodes;
const createSchemaCustomization = (args, options) => __awaiter(void 0, void 0, void 0, function* () {
    if (!(0, utils_1.validOptions)(options)) {
        return;
    }
    const executor = (0, create_query_executor_1.createQueryExecutor)(options);
    // TODO: Accept configuration and fragment overrides from plugin settings.
    const config = yield (0, create_sourcing_config_1.createSourcingConfig)(args, executor, options);
    yield (0, gatsby_graphql_source_toolkit_1.createSchemaCustomization)(config);
    yield (0, create_translation_query_field_1.createTranslationQueryField)(args, (0, create_query_executor_1.createQueryExecutor)(options));
    args.actions.createTypes(`
    type Query {
      drupalBuildId: Int!
    }
  `);
});
exports.createSchemaCustomization = createSchemaCustomization;
const createPages = (args, options) => __awaiter(void 0, void 0, void 0, function* () {
    yield (0, create_pages_1.createPages)(args, options);
    const buildId = yield args.cache.get(`LAST_BUILD_ID_TMP`);
    yield args.cache.set('LAST_BUILD_ID', buildId);
});
exports.createPages = createPages;
const createResolvers = ({ createResolvers, cache, }) => __awaiter(void 0, void 0, void 0, function* () {
    createResolvers({
        Query: {
            drupalBuildId: {
                resolve() {
                    return __awaiter(this, void 0, void 0, function* () {
                        return (yield cache.get(`LAST_BUILD_ID`)) || -1;
                    });
                },
            },
        },
    });
});
exports.createResolvers = createResolvers;
const onPostBuild = (args) => __awaiter(void 0, void 0, void 0, function* () {
    const data = {
        drupalBuildId: (yield args.cache.get('LAST_BUILD_ID')) || -1,
    };
    const path = `${args.store.getState().program.directory}/public/build.json`;
    args.reporter.info(`Writing build information to ${path}.`);
    fs_1.default.writeFileSync(path, JSON.stringify(data, null, 2));
});
exports.onPostBuild = onPostBuild;
